/*******************************************************************************
 * main.cpp - Contains initilization, window management, file input
 *            and fragment shader code
 *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <time.h>
#include <string>

#include "util.h"
#include "sat.h"
#include "painter.h"
#include "animator.h"
#include "algebra.h"

using namespace std;

// Function pointers for OpenGL extensions
#ifdef _WIN32
PFNGLGENFRAGMENTSHADERSATIPROC glGenFragmentShadersATI;
PFNGLBINDFRAGMENTSHADERATIPROC glBindFragmentShaderATI;
PFNGLBEGINFRAGMENTSHADERATIPROC glBeginFragmentShaderATI;
PFNGLENDFRAGMENTSHADERATIPROC glEndFragmentShaderATI;
PFNGLPASSTEXCOORDATIPROC glPassTexCoordATI;
PFNGLCOLORFRAGMENTOP1ATIPROC glColorFragmentOp1ATI;
PFNGLCOLORFRAGMENTOP2ATIPROC glColorFragmentOp2ATI;
PFNGLCOLORFRAGMENTOP3ATIPROC glColorFragmentOp3ATI;
PFNGLALPHAFRAGMENTOP1ATIPROC glAlphaFragmentOp1ATI;
PFNGLALPHAFRAGMENTOP2ATIPROC glAlphaFragmentOp2ATI;
PFNGLALPHAFRAGMENTOP3ATIPROC glAlphaFragmentOp3ATI;
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB;
PFNGLSETFRAGMENTSHADERCONSTANTATIPROC glSetFragmentShaderConstantATI;
PFNGLSAMPLEMAPATIPROC glSampleMapATI;
#endif

//MODE SETTINGS
bool BMPRender = false;
GLuint texname = 0;
bool forcestart = false;
bool ModelTexture = true;
bool ModeDemo = false;
bool ModeOutput = false;

// Global variables because this is ugly and should be contained in a class
Point3D camerapos(0, 0, 20);
string outfilename;
GLuint particletex = 0;
int MDstage = 0;
int MDstageprev = 0;
int VideoFlags = 0;                                        // Video Flags for the Create Window function
SDL_Surface * MainWindow = NULL;                           // drawing surface on the SDL window
Painter canvas;
Animator anim;
//SDL_Surface * image = NULL;
//rgb * pic = NULL;
int radius = 0;
int frames = 0;
int ScreenWidth = 0;
int ScreenHeight = 0;
bool filter = true;
bool ps = false;

// If not under windows, permanently disable the pixel shaders
#ifdef _WIN32
bool hasps = true;
#else
bool hasps = false;
#endif

bool buff = true;
rgb bgcolor;
rgb * pic = NULL;
rgb * savebuff = NULL;
rgb * buff1 = NULL;
rgb * buff2 = NULL;
rgba * b1 = NULL;
rgba * b2 = NULL;
rgba * b3 = NULL;
int starttick = 0;
int pausetick = 0;
GLuint blurtex[4];
GLuint blurfilter;
GLuint difffilter;

// Perform memory cleanup then quit the program
void Quit( int code ) {
	if( pic != NULL ) {
		delete [] pic;
	}
	if( savebuff != NULL ) {
		delete [] savebuff;
	}
	if( buff1 != NULL ) {
		delete [] buff1;
	}
	if( buff2 != NULL ) {
		delete [] buff2;
	}
	if( b1 != NULL ) {
		delete [] b1;
	}
	if( b2 != NULL ) {
		delete [] b2;
	}
	if( b3 != NULL ) {
		delete [] b3;
	}
	SDL_Quit();
	exit( code );
}

// Create an OpenGL Window
void CreateGLWindow(const char * strWindowName, int width, int height, int VideoFlags)
{
    MainWindow = SDL_SetVideoMode(width, height, SCREEN_DEPTH, VideoFlags);

    if( MainWindow == NULL )                                           
    {
        cerr << "Failed to Create Window : " << SDL_GetError() << endl;
        Quit(0);
    }

    SDL_WM_SetCaption(strWindowName, strWindowName); 
}

// Set window formating options for OpenGL with hardware acceleration
void SetWindowFormat(void)
{
    
    VideoFlags	= SDL_OPENGL;                            
    VideoFlags	|= SDL_HWPALETTE;  

    const SDL_VideoInfo * VideoInfo = SDL_GetVideoInfo(); 

    if(VideoInfo == NULL) 
    {
        cerr << "Failed getting Video Info : " << SDL_GetError() << endl;  
        Quit(0);
    }

    VideoFlags |= SDL_HWSURFACE;
    
    if(VideoInfo -> blit_hw)      
        VideoFlags |= SDL_HWACCEL;

    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE,   SCREEN_DEPTH);
    SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 0);
    SDL_GL_SetAttribute( SDL_GL_ACCUM_RED_SIZE, 0); 
    SDL_GL_SetAttribute( SDL_GL_ACCUM_GREEN_SIZE, 0); 
    SDL_GL_SetAttribute( SDL_GL_ACCUM_BLUE_SIZE, 0);
    SDL_GL_SetAttribute( SDL_GL_ACCUM_ALPHA_SIZE, 0);
	
}

// Initilize the OpenGL Settings
void Initialize( int w, int h )
{
	ScreenWidth = w;
	ScreenHeight = h;

	// Clear the buffers and set them to the appropriate background colour
	SetWindowFormat();
	CreateGLWindow( "Oil Painter", w, h, VideoFlags );
	glReadBuffer( GL_BACK );
	glDrawBuffer( GL_FRONT );
	glClearColor((float)bgcolor.r/255, (float)bgcolor.g/255, (float)bgcolor.b/255, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawBuffer( GL_BACK );
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Create buffers for storing screen information
	buff1 = new rgb[w*h];
	buff2 = new rgb[w*h];
	savebuff = new rgb[w*h];
	b1 = new rgba[w*h];
	b2 = new rgba[w*h];
	b3 = new rgba[w*h];

	//glEnable( GL_TEXTURE_2D );
	glEnable( GL_COLOR_MATERIAL );
	glShadeModel(GL_SMOOTH);
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	glPixelStorei( GL_PACK_ALIGNMENT, 1 );

    glViewport(0,0,w,h);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

	// Single Lightsource at the camera
	GLfloat light_ambient[] = { 0.2f, 0.2f, 0.2f, 1.0f };
	GLfloat light_diffuse[] = { 0.5f, 0.5f, 0.5f, 1.0f };
	GLfloat light_specular[] = { 0.5f, 0.5f, 0.5f, 1.0f };
	GLfloat light1_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };

	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_POSITION, light1_position);
	glEnable(GL_LIGHT0);

	// Used for drawing single dots on the screen for testing purposes
	glNewList( 8, GL_COMPILE );
	gluDisk( gluNewQuadric(), 0, 8, 32, 1 );
	glEndList();
	glNewList( 4, GL_COMPILE );
	gluDisk( gluNewQuadric(), 0, 4, 16, 1 );
	glEndList();
	glNewList( 2, GL_COMPILE );
	gluDisk( gluNewQuadric(), 0, 2, 8, 1 );
	glEndList();
}

// Main render pipeline
void render( int tick )
{
	// Bitmap Render Mode
	if( BMPRender ) {
		if( tick == 0 ) {
			if( !ModeDemo || MDstage < 4 )
				glClear(GL_COLOR_BUFFER_BIT );

			  glMatrixMode(GL_PROJECTION);
			  glLoadIdentity();
			  glOrtho( 0, ScreenWidth-1, 0, ScreenHeight-1, 0, 1 );
			  glMatrixMode(GL_MODELVIEW);

			  glDisable(GL_LIGHTING);
			  glDisable(GL_DEPTH_TEST);
			  glDisable(GL_NORMALIZE);
			  glDisable(GL_CULL_FACE);

			  // Run the painting algorithm
			  canvas.paint( pic, true );
		}
	// Animation Mode
	} else if( tick >= 0 ) {

		// Setup OpenGL for 3D animation rendering
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f,(GLfloat)ScreenWidth/(GLfloat)ScreenHeight, 1 ,150.0f);

		glEnable(GL_LIGHTING);
		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_NORMALIZE);
		glEnable(GL_CULL_FACE);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// Translate by the camera position
		glTranslated( camerapos[0], camerapos[1], -camerapos[2] );

		// Render the current animation Frame
		anim.render( tick, ModelTexture );
		
		glPopMatrix();
		glFlush();

		// Check for Errors
		GLenum error = glGetError();
		if( error != GL_NO_ERROR ) {
			cerr << "ERROR!! " << ++frames << endl;
			Quit(0);
		}
		// If set to apply the Oil Painting Filter then do so
		if( filter && (!ModeDemo || MDstage != 0) ) {
			// Copy the current frame to main memory to perform filtering
			if( buff ) {
				glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGB, GL_UNSIGNED_BYTE, (void *)buff1);
			} else {
				glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGB, GL_UNSIGNED_BYTE, (void *)buff2);
			}
#ifdef _WIN32
		// Using the fragment shaders to perform the blur operations
			if( ps && hasps ) {
				// save the current image as a texture then blend slightly offset
				// textures to simulate a gaussian blur
				glPushAttrib(GL_ALL_ATTRIB_BITS );
				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[0]);
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );

				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[0]);

				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[0]);

				glActiveTextureARB(GL_TEXTURE3_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[0]);

				glActiveTextureARB(GL_TEXTURE4_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[0]);

				glDisable(GL_TEXTURE_GEN_S);
				glDisable(GL_TEXTURE_GEN_T);
				glDisable(GL_LIGHTING);
				glEnable(GL_BLEND);
				glDisable(GL_DEPTH_TEST);	

				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glOrtho( 0, ScreenWidth, 0, ScreenHeight, 0, 1 );
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();
				glEnable(GL_FRAGMENT_SHADER_ATI);
				glBindFragmentShaderATI( blurfilter );

				glCallList(31);

				glFlush();

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[1]);
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );
				
				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[1]);

				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[1]);

				glActiveTextureARB(GL_TEXTURE3_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[1]);

				glActiveTextureARB(GL_TEXTURE4_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[1]);

				glCallList( 32 );
				glFlush();
				// Save First Blur Image
				glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGBA, GL_UNSIGNED_BYTE, (void *)b1);

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[1]);
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );

				glCallList( 33 );
				glFlush();

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[2]);
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );

				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[2]);

				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[2]);

				glActiveTextureARB(GL_TEXTURE3_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[2]);

				glActiveTextureARB(GL_TEXTURE4_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[2]);

				glCallList( 34 );

				glFlush();
				// Save Second Blur Image
				glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGBA, GL_UNSIGNED_BYTE, (void *)b2);

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[2]);
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );

				glCallList( 31 );

				glFlush();

				glActiveTextureARB(GL_TEXTURE0_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[3]);
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );

				glActiveTextureARB(GL_TEXTURE1_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[3]);

				glActiveTextureARB(GL_TEXTURE2_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[3]);

				glActiveTextureARB(GL_TEXTURE3_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[3]);

				glActiveTextureARB(GL_TEXTURE4_ARB);
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, blurtex[3]);

				glCallList( 32 );


				glFlush();
				// Save Third Blur Image
				glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );
				glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGBA, GL_UNSIGNED_BYTE, (void *)b3);

				glDisable( GL_FRAGMENT_SHADER_ATI );
				
				glPopAttrib();
			}
#endif // _WIN32  Fragment Shader section

			// If starting fresh clear the buffer
			if( !forcestart ) {
				glReadBuffer(GL_FRONT);
				glCopyPixels( 0, 0, ScreenWidth, ScreenHeight, GL_COLOR );
				glReadBuffer(GL_BACK);
			} else {
				glClear( GL_COLOR_BUFFER_BIT );
			}

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();

			glDisable(GL_LIGHTING);
			glEnable( GL_BLEND );
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_NORMALIZE);
			glDisable(GL_CULL_FACE);

			glOrtho( 0, ScreenWidth, 0, ScreenHeight, 0, 1 );
			glMatrixMode(GL_MODELVIEW);
			bool start = (tick == 0)?true:false;
			start = start || forcestart;
			// Run the painting algorithm
			if( buff ) {
				if( ps && hasps ) {
					canvas.paint( buff1, b1, b2, b3, start );
				} else {
					canvas.paint( buff1, start );
				}
			} else {
				if( ps && hasps ) {
					canvas.paint( buff2, b1, b2, b3, start );
				} else {
					canvas.paint( buff2, start );
				}
			}
			buff = !buff;
		}

		frames++;
	}

  SDL_GL_SwapBuffers();
  
  // Write images to file
	if( ModeOutput ) {
		glReadBuffer( GL_FRONT );
		glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGB, GL_UNSIGNED_BYTE, (void *)savebuff);
		glReadBuffer( GL_BACK );
		SDL_Surface *image = SDL_CreateRGBSurface(SDL_SWSURFACE, 
				ScreenWidth, ScreenHeight, 24, 
				MainWindow->format->Rmask, MainWindow->format->Gmask, 
				MainWindow->format->Bmask, MainWindow->format->Amask);
		convrgb( image, savebuff );

		if( BMPRender ) {
			SDL_SaveBMP( image, outfilename.c_str() );
		} else {
			char num[5];
			sprintf( num, "%04d", tick/100 );
			num[4] = 0;
			string numfilename( outfilename );
			numfilename.insert( numfilename.find('.') , num );
			SDL_SaveBMP( image, numfilename.c_str() );
		}

		SDL_FreeSurface( image );
			
	}
}

// Update the current pause point
void updatepause(){
	if( pausetick != 0 ) {
		if( starttick > pausetick ) starttick = pausetick-1;
		render( pausetick-starttick );
	}
}

// Manages keyboard inputs
bool KeyPressHandler( SDL_keysym * keysym )
{
	switch( keysym->sym ) {
		// ESCAPE - exits program
		case SDLK_ESCAPE:
			return true;
		// F - toggles the painting filter
		case SDLK_f:
			filter = !filter;
			forcestart = true;
			updatepause();
			break;
		// O - Toggle fragment shader
		case SDLK_o:
			if( !ModeDemo ) ps = !ps;
			updatepause();
			break;
		// R - Restart Animation
		case SDLK_r:
			starttick = SDL_GetTicks();
			if( pausetick != 0 ) {
				pausetick = starttick + 1;
			}
			if( ModeDemo ) MDstageprev = 0;
			render(0);
			break;
		// P - play/pause
		case SDLK_p:
			if( pausetick == 0 ) {
				pausetick = SDL_GetTicks();
			} else {
				starttick = SDL_GetTicks()-(pausetick - starttick);
				pausetick = 0;
			}
			break;
		// B - Toggle brush textures
		case SDLK_b:
			canvas.togletexture();
			if( BMPRender ) {
				if( ModeDemo ) MDstageprev = 0;
				render(0);
			} else {
				updatepause();
			}
			break;
		// H, J, K, L - Fast rewind, rewind, forward, fast forward
		case SDLK_k:
			starttick -= 100;
			updatepause();
			break;
		case SDLK_j:
			starttick += 100;
			if( starttick > SDL_GetTicks() ) starttick = SDL_GetTicks();
			updatepause();
			break;
		case SDLK_l:
			starttick -= 1000;
			updatepause();
			break;
		case SDLK_h:
			starttick += 1000;
			if( starttick > SDL_GetTicks() ) starttick = SDL_GetTicks();
			updatepause();
			break;
		// T - Toggle model textures
		case SDLK_t:
			ModelTexture = !ModelTexture;
			updatepause();
			break;
		// 0-9 changing Demo Modes
		default:
			if( keysym->sym >= SDLK_0 && keysym->sym <= SDLK_9 ) {
				if( ModeDemo ) {
					MDstageprev = MDstage;
					MDstage = keysym->sym - SDLK_0;
					if( BMPRender ) render(0);
					if( pausetick != 0) updatepause();
				}
			}
			break;
	}
	return false;
}

// Main program operation loop
void MainLoop()
{
	bool done = false;
	SDL_Event event;
	int ticks = SDL_GetTicks();
	render( 0 );
	starttick = SDL_GetTicks();
	
	// If a Bitmap and in Output Mode, program quits after rendering
	if( ModeOutput ) {
		if( BMPRender ) done = true;
		starttick = 0;
	}
	
	// Loop that handles inputs and rendering
	while(!done) {
		while( SDL_PollEvent( &event ) ) {
			switch( event.type ) {
				case SDL_QUIT:
					done = true;
					break;

				case SDL_KEYDOWN:
					if( ModeOutput ) {
						if( event.key.keysym.sym == SDLK_ESCAPE ) done = true;	
					} else {
						done = KeyPressHandler( &event.key.keysym );
					}
					break;

				case SDL_VIDEOEXPOSE:
					if( !ModeOutput ) {
						if( BMPRender ) {
							render(1);
						} else if( pausetick != 0 ){
							render(-1);
						}
					}

				default:
					break;
			}
		}
		// If in Output mode and Animating, program ends after animation complete
		if( ModeOutput ) {
			starttick += 100;
			render( starttick );
			if( starttick >= anim.getLength() ) done = true;
		} else if( pausetick == 0 ) {
			render(SDL_GetTicks() - starttick);
		}
		forcestart = false;
	}
	int seconds = (SDL_GetTicks() - ticks) /1000;
	if( seconds == 0 ) seconds = 1;
	cerr << "FPS: " << (float)frames/seconds << endl;
	cerr << "Total Frames: " << frames << endl;
}

// OpenGL initilization for particle system
void initparticlelist( const char* pbmp) {
	SDL_Surface *teximg;
	//GLuint particletex;
	teximg = SDL_LoadBMP(pbmp);
	if( teximg == NULL ) {
		cerr << "Couldn't load texture BMP" << endl;
		Quit(0);
	}
	glGenTextures(1, &particletex);
	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, particletex );
	Uint8 *tex = new Uint8[teximg->w*teximg->h];
	convalpha( teximg, tex );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA, teximg->w, teximg->h, 0, GL_ALPHA, GL_UNSIGNED_BYTE, tex );
	delete [] tex;
	SDL_FreeSurface( teximg );

	glPushAttrib( GL_ALL_ATTRIB_BITS );
		glNewList( particlelist, GL_COMPILE );
		glEnable( GL_BLEND );
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		glEnable( GL_TEXTURE_2D );

		glBindTexture( GL_TEXTURE_2D, particletex );
		glBegin(GL_TRIANGLE_STRIP);
		glNormal3d(0, 0, 1);
		glTexCoord2f(1,1); glVertex3f(0.2f,0.2f,0);
		glTexCoord2f(0,1); glVertex3f(-0.2f,0.2f,0);
		glTexCoord2f(1,0); glVertex3f(0.2f,-0.2f,0);
		glTexCoord2f(0,0); glVertex3f(-0.2f,-0.2f,0);
		glEnd();
		glEndList();
	glPopAttrib();
}

// Copies current canvas, used for fragment shaders
void getcanvas( rgba*c, int r ) {
#ifdef _WIN32
	glPushAttrib(GL_ALL_ATTRIB_BITS );

	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, blurtex[3-r]);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, blurtex[0]);
	glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, ScreenWidth, ScreenHeight, 0 );

	glEnable(GL_FRAGMENT_SHADER_ATI);
	glBindFragmentShaderATI( difffilter );
	glCallList(37);
	glFlush();
	glReadPixels( 0, 0, ScreenWidth, ScreenHeight, GL_RGBA, GL_UNSIGNED_BYTE, (void *)c);

	glDisable(GL_FRAGMENT_SHADER_ATI);
	glPopAttrib();
#endif
}

// Fragment shader initilization code
void initblurtextures( int w, int h ){
#ifdef _WIN32
  // Declare function pointers
	glGenFragmentShadersATI = (PFNGLGENFRAGMENTSHADERSATIPROC) wglGetProcAddress("glGenFragmentShadersATI");
	glBindFragmentShaderATI = (PFNGLBINDFRAGMENTSHADERATIPROC) wglGetProcAddress("glBindFragmentShaderATI");
	glBeginFragmentShaderATI = (PFNGLBEGINFRAGMENTSHADERATIPROC) wglGetProcAddress("glBeginFragmentShaderATI");
	glEndFragmentShaderATI = (PFNGLENDFRAGMENTSHADERATIPROC) wglGetProcAddress("glEndFragmentShaderATI");
	glPassTexCoordATI = (PFNGLPASSTEXCOORDATIPROC) wglGetProcAddress("glPassTexCoordATI");
	glColorFragmentOp1ATI = (PFNGLCOLORFRAGMENTOP1ATIPROC) wglGetProcAddress("glColorFragmentOp1ATI");
	glColorFragmentOp2ATI = (PFNGLCOLORFRAGMENTOP2ATIPROC) wglGetProcAddress("glColorFragmentOp2ATI");
	glColorFragmentOp3ATI = (PFNGLCOLORFRAGMENTOP3ATIPROC) wglGetProcAddress("glColorFragmentOp3ATI");
	glAlphaFragmentOp1ATI = (PFNGLALPHAFRAGMENTOP1ATIPROC) wglGetProcAddress("glAlphaFragmentOp1ATI");
	glAlphaFragmentOp2ATI = (PFNGLALPHAFRAGMENTOP2ATIPROC) wglGetProcAddress("glAlphaFragmentOp2ATI");
	glAlphaFragmentOp3ATI = (PFNGLALPHAFRAGMENTOP3ATIPROC) wglGetProcAddress("glAlphaFragmentOp3ATI");
	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC) wglGetProcAddress("glActiveTextureARB");
	glMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC) wglGetProcAddress("glMultiTexCoord2fARB");
	glSetFragmentShaderConstantATI = (PFNGLSETFRAGMENTSHADERCONSTANTATIPROC) wglGetProcAddress("glSetFragmentShaderConstantATI");
	glSampleMapATI = (PFNGLSAMPLEMAPATIPROC) wglGetProcAddress("glSampleMapATI");
	if( glBeginFragmentShaderATI == NULL ) {
		hasps = false;
		return;
	}
	Uint8 *blank;
	blank = new Uint8[w*h*4];
	memset(blank, 0, 4*w*h);
	glGenTextures(3, &blurtex[0]);
	for( int i = 0; i < 4; i++ ) {
		glEnable( GL_TEXTURE_2D );
		glBindTexture( GL_TEXTURE_2D, blurtex[i] );
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, blank );
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}
	delete [] blank;
	
	glEnable( GL_FRAGMENT_SHADER_ATI );
	glEnable( GL_TEXTURE_2D );
	blurfilter = glGenFragmentShadersATI( 1 );
	GLfloat fdiv5[4] = { 0.2f, 0.2f, 0.2f, 0.0 };
	GLfloat flum[4] = { 0.3f, 0.59f, 0.11f, 0.0 };
	glBindFragmentShaderATI( blurfilter );
	glBeginFragmentShaderATI();
	glSetFragmentShaderConstantATI( GL_CON_0_ATI, fdiv5 );
	glSetFragmentShaderConstantATI( GL_CON_1_ATI, flum );

	glSampleMapATI( GL_REG_0_ATI, GL_TEXTURE0_ARB, GL_SWIZZLE_STR_ATI); 
	glSampleMapATI( GL_REG_1_ATI, GL_TEXTURE1_ARB, GL_SWIZZLE_STR_ATI); 
	glSampleMapATI( GL_REG_2_ATI, GL_TEXTURE2_ARB, GL_SWIZZLE_STR_ATI); 
	glSampleMapATI( GL_REG_3_ATI, GL_TEXTURE3_ARB, GL_SWIZZLE_STR_ATI); 
	glSampleMapATI( GL_REG_4_ATI, GL_TEXTURE4_ARB, GL_SWIZZLE_STR_ATI); 
	glColorFragmentOp2ATI( GL_ADD_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_1_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp2ATI( GL_ADD_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_2_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp2ATI( GL_ADD_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_3_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp2ATI( GL_ADD_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_4_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp2ATI( GL_MUL_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_CON_0_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp2ATI( GL_MUL_ATI, GL_REG_2_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_CON_1_ATI, GL_NONE, GL_NONE );
	// Can't seem to find a no-op for Color Fragment
	glColorFragmentOp1ATI( GL_MOV_ATI, GL_REG_1_ATI, GL_NONE, GL_NONE,
						GL_REG_1_ATI, GL_NONE, GL_NONE );
	glAlphaFragmentOp2ATI( GL_ADD_ATI, GL_REG_0_ATI, GL_NONE, 
						GL_REG_2_ATI, GL_RED, GL_NONE,
						GL_REG_2_ATI, GL_BLUE, GL_NONE );
	glAlphaFragmentOp2ATI( GL_ADD_ATI, GL_REG_0_ATI, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_2_ATI, GL_BLUE, GL_NONE );
	glEndFragmentShaderATI();

	difffilter = glGenFragmentShadersATI( 1 );
	GLfloat fdiv6[4] = { 0.166f, 0.166f, 0.166f, 0.166f };
	GLfloat fzero[4] = { 1.0f, 1.0f, 1.0f, 1.0f };

	glBindFragmentShaderATI( difffilter );
	glBeginFragmentShaderATI();
	glSetFragmentShaderConstantATI( GL_CON_0_ATI, fdiv6 );
	glSampleMapATI( GL_REG_0_ATI, GL_TEXTURE0_ARB, GL_SWIZZLE_STR_ATI); 
	glSampleMapATI( GL_REG_1_ATI, GL_TEXTURE1_ARB, GL_SWIZZLE_STR_ATI); 
	glColorFragmentOp2ATI( GL_SUB_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_1_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp2ATI( GL_DOT3_ATI, GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_0_ATI, GL_NONE, GL_NONE );
	glAlphaFragmentOp2ATI( GL_DOT3_ATI, GL_REG_0_ATI, GL_NONE, 
						GL_REG_0_ATI, GL_NONE, GL_NONE,
						GL_REG_0_ATI, GL_NONE, GL_NONE );
	glColorFragmentOp1ATI( GL_MOV_ATI, GL_REG_0_ATI, GL_RED_BIT_ATI|GL_GREEN_BIT_ATI|GL_BLUE_BIT_ATI, GL_NONE,
						GL_REG_1_ATI, GL_NONE, GL_NONE );
	
	glEndFragmentShaderATI();
	GLuint error = glGetError();
	if( error != GL_NO_ERROR ) {
		cerr << "ERROR IN SHADER" << endl;
	}

	glDisable(GL_FRAGMENT_SHADER_ATI);
	float wf = (float)w;
	float hf = (float)h;
	glNewList( 31, GL_COMPILE );
		glBegin(GL_POLYGON);
		//glTexCoord2f(0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 2.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 3.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 4.0f/(ScreenWidth),0); 
		glVertex3f(2,0,0);
		//glTexCoord2f(1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f-4.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f-3.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1.0f-2.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1.0f-1.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1.0f,0); 
		glVertex3f(ScreenWidth-2,0,0);
		//glTexCoord2f(1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f-4.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f-3.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1.0f-2.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1.0f-1.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1.0f,1); 
		glVertex3f(ScreenWidth-2,ScreenHeight,0);
		//glTexCoord2f(0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 2.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 3.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 4.0f/(ScreenWidth),1); 
		glVertex3f(2,ScreenHeight,0);
		glEnd();
	glEndList();

	glNewList( 32, GL_COMPILE );
		glBegin( GL_POLYGON );
		glNormal3d(0, 0, -1);
		//glTexCoord2f(0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,1.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 0,2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 0,3.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 0,4.0f/(ScreenHeight)); 
		glVertex3f(0,2,0);
		//glTexCoord2f(1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,1.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1,2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1,3.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1,4.0f/(ScreenHeight)); 
		glVertex3f(ScreenWidth,2,0);
		//glTexCoord2f(1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,1.0f-4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,1.0f-3.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1,1.0f-2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1,1.0f-1.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1,1.0f); 
		glVertex3f(ScreenWidth,ScreenHeight-2,0);
		//glTexCoord2f(0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1.0f-4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,1.0f-3.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 0,1.0f-2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 0,1.0f-1.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 0,1.0f); 
		glVertex3f(0,ScreenHeight-2,0);
		glEnd();
	glEndList();
		
	glNewList( 33, GL_COMPILE );
		glBegin(GL_POLYGON);
		//glTexCoord2f(0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 4.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 6.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 8.0f/(ScreenWidth),0); 
		glVertex3f(4,0,0);
		//glTexCoord2f(1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f-8.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f-6.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1.0f-4.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1.0f-2.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1.0f,0); 
		glVertex3f(ScreenWidth-4,0,0);
		//glTexCoord2f(1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f-8.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f-6.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1.0f-4.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1.0f-2.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1.0f,1); 
		glVertex3f(ScreenWidth-4,ScreenHeight,0);
		//glTexCoord2f(0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 2.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 4.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 6.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 8.0f/(ScreenWidth),1); 
		glVertex3f(4,ScreenHeight,0);
		glEnd();
	glEndList();

	glNewList( 34, GL_COMPILE );
		glBegin( GL_POLYGON );
		glNormal3d(0, 0, -1);
		//glTexCoord2f(0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 0,4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 0,6.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 0,8.0f/(ScreenHeight)); 
		glVertex3f(0,4,0);
		//glTexCoord2f(1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1,4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1,6.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1,8.0f/(ScreenHeight)); 
		glVertex3f(ScreenWidth,4,0);
		//glTexCoord2f(1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,1.0f-8.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,1.0f-6.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1,1.0f-4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1,1.0f-2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1,1.0f); 
		glVertex3f(ScreenWidth,ScreenHeight-4,0);
		//glTexCoord2f(0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1.0f-8.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,1.0f-6.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 0,1.0f-4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 0,1.0f-2.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 0,1.0f); 
		glVertex3f(0,ScreenHeight-4,0);
		glEnd();
	glEndList();

	glNewList( 35, GL_COMPILE );
		glBegin(GL_POLYGON);
		//glTexCoord2f(0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 4.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 8.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 12.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 16.0f/(ScreenWidth),0); 
		glVertex3f(8,0,0);
		//glTexCoord2f(1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f-16.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f-12.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1.0f-8.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1.0f-4.0f/(ScreenWidth),0); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1.0f,0); 
		glVertex3f(ScreenWidth-8,0,0);
		//glTexCoord2f(1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f-16.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1.0f-12.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1.0f-8.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1.0f-4.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1.0f,1); 
		glVertex3f(ScreenWidth-8,ScreenHeight,0);
		//glTexCoord2f(0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 4.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 8.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 12.0f/(ScreenWidth),1); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 16.0f/(ScreenWidth),1); 
		glVertex3f(8,ScreenHeight,0);
		glEnd();
	glEndList();

	glNewList( 36, GL_COMPILE );
		glBegin( GL_POLYGON );
		glNormal3d(0, 0, -1);
		//glTexCoord2f(0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 0,8.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 0,12.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 0,16.0f/(ScreenHeight)); 
		glVertex3f(0,8,0);
		//glTexCoord2f(1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1,8.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1,12.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1,16.0f/(ScreenHeight)); 
		glVertex3f(ScreenWidth,8,0);
		//glTexCoord2f(1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,1.0f-16.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,1.0f-12.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 1,1.0f-8.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 1,1.0f-4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 1,1.0f); 
		glVertex3f(ScreenWidth,ScreenHeight-8,0);
		//glTexCoord2f(0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1.0f-16.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,1.0f-12.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE2_ARB, 0,1.0f-8.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE3_ARB, 0,1.0f-4.0f/(ScreenHeight)); 
		glMultiTexCoord2fARB(GL_TEXTURE4_ARB, 0,1.0f); 
		glVertex3f(0,ScreenHeight-8,0);
		glEnd();
	glEndList();

	glNewList( 37, GL_COMPILE );
		glBegin( GL_POLYGON );
		glNormal3d(0, 0, -1);
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,0); 
		glVertex3f(0,0,0);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,0); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,0); 
		glVertex3f(ScreenWidth,0,0);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1,1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1,1); 
		glVertex3f(ScreenWidth,ScreenHeight,0);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0,1); 
		glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0,1); 
		glVertex3f(0,ScreenHeight,0);
		glEnd();
	glEndList();
#endif
}

// Load a bitmap from a file
bool loadBMP( const char* filename, int &width, int &height ) {
	SDL_Surface *image;
	image = SDL_LoadBMP(filename);
	if( image == NULL ) {
		cerr << "Couldn't load BMP" << endl;
		Quit(0);
	}
	width = image->w;
	height = image->h;
	pic = new rgb[image->h*image->w];
	convscene( image, pic );
    SDL_FreeSurface( image );
	return true;
}

// Load a brush texture from a file
bool loadBrush( const char* filename ) {
	SDL_Surface * teximg;

	teximg = SDL_LoadBMP(filename);
	if( teximg == NULL ) {
		cerr << "Couldn't load brush BMP" << endl;
		Quit(0);
	}
	
	glGenTextures(1, &texname);
	glBindTexture( GL_TEXTURE_2D, texname );
	Uint8 *tex = new Uint8[teximg->w*teximg->h];
	convalpha( teximg, tex );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA, teximg->w, teximg->h, 0, GL_ALPHA, GL_UNSIGNED_BYTE, tex );
	canvas.setbrushtexture( texname );
	delete [] tex;
	SDL_FreeSurface( teximg );
	return true;
}

// Load XML Scene File
bool loadXML( char* filename ) {
	int w, h;
	TiXmlDocument doc;
	TiXmlElement* element = NULL;
	const char *particlename = NULL;
	if( !doc.LoadFile( filename ) ) {
		return false;
	}
	// Read Painter information from XML file
	element = doc.FirstChildElement( "painter" ); 
	if( element == NULL ) return false;
	TiXmlElement* bmpElem = NULL;
	bmpElem = element->FirstChildElement("bmp");
	// Read Bitmap information and get height and width
	if( bmpElem != NULL ) {
		BMPRender = true;
		ps = false;
		TiXmlText* bmpText = bmpElem->FirstChild()->ToText();
		loadBMP( bmpText->Value(), w, h );
	} else {
	// If in animation mode, height and width is specified in the Scene file
		TiXmlElement* resElem = NULL;
		resElem = element->FirstChildElement("resolution");
		TiXmlText* resText = resElem->FirstChild()->ToText();
		sscanf( resText->Value(), "%d,%d", &w, &h);
	}
	
	// Read the background colour
	int r, g, b;
	TiXmlElement* colElem = NULL;
	colElem = element->FirstChildElement("bgcolor");
	if( colElem != NULL ) {
		TiXmlText* colText = colElem->FirstChild()->ToText();
		sscanf( colText->Value(), "%d,%d,%d", &r, &g, &b);
		bgcolor.r = (Uint8)r;
		bgcolor.g = (Uint8)g;
		bgcolor.b = (Uint8)b;
	}
	// Check for output Mode
	TiXmlElement* outElem = NULL;
	outElem = element->FirstChildElement("output");
	if( outElem != NULL ) {
		TiXmlText* outText = outElem->FirstChild()->ToText();
		outfilename = outText->Value();
		ModeOutput = true;
	} else {
		// Check for Demo Mode
		TiXmlElement* demoElem = NULL;
		demoElem = element->FirstChildElement("demo");
		if( demoElem != NULL ) {
			ModeDemo = true;
			MDstage = 6;
		} else {
			TiXmlElement* psElem = NULL;
			psElem = element->FirstChildElement("ps");
			if( psElem != NULL ) {
				ps = true;
			}
		}
	}
	// Read in Brush parameters
	TiXmlElement* bparamElem = NULL;
	bparamElem = element->FirstChildElement("bparam");
	if( bparamElem != NULL ) {
		Uint32 T;
		int maxlen, minlen;
		TiXmlText* bparamText = bparamElem->FirstChild()->ToText();
		sscanf( bparamText->Value(), "%u,%d,%d", &T, &maxlen, &minlen);
		canvas.setstrokeparam( T, maxlen, minlen );
	}
	// Read Camera Position
	TiXmlElement* posElem = element->FirstChildElement( "cpos" );
	if( posElem ) {
		TiXmlText* posText = posElem->FirstChild()->ToText();
		sscanf( posText->Value(), "%le,%le,%le", &camerapos[0], &camerapos[1], &camerapos[2]);
	}

	// Initilize the window
	Initialize( w, h );

	// Read in texture files
	TiXmlElement* texElem = NULL;
	texElem = element->FirstChildElement("btex");
	if( texElem != NULL ) {
		TiXmlText* texText = texElem->FirstChild()->ToText();
		loadBrush( texText->Value() );
	}

	TiXmlElement* parElem = NULL;
	parElem = element->FirstChildElement("particle");
	if( parElem != NULL ) {
		TiXmlText* parText = parElem->FirstChild()->ToText();
		particlename = parText->Value();
	}
	
	// If not working in a bitmap image, read in animation information
	if( !BMPRender ) {
		anim.loadScene( &doc );
		if( particlename == NULL ) {
			initparticlelist( "particle.bmp" );
		} else {
			initparticlelist( particlename );
		}
		initblurtextures( ScreenWidth, ScreenHeight );
	}
	canvas.setsize( ScreenWidth, ScreenHeight );
	return true;
}

int main(int argc, char **argv)
{
	srand( (unsigned)time( NULL ) );

	if( argc < 2 ) {
		cerr << "Usage:\n\t" << argv[0] << " <xml scene file>" << endl;
		exit(0);
	}

#ifdef DEBUG
	if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE ) < 0 )
#else
	if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
#endif
	{
		cerr << "Video initialization failed: "  <<	SDL_GetError( ) << endl;
		Quit( 1 );
	}

	if( !loadXML( argv[1] ) ) {
		cerr << "Invalid XML file: " << argv[1] << endl;
		exit(0);
	}


	MainLoop();
	Quit(0);
	return 0;
}
