/********
Model Class
*********/
#include "material.h"
#include "util.h"

pMaterial::pMaterial( const char *mname ):
name( mname ) {
	inittextures();
	diffuse = RGB_0;
	specular = RGB_0;
}

pMaterial::pMaterial( const char *mname, rgb d ):
name( mname ) {
	inittextures();
	diffuse = d;
	specular = RGB_0;
}

pMaterial::pMaterial( const char *mname, rgb d, rgb s ):
name( mname) {
	inittextures();
	diffuse = d;
	specular = s;
}

void pMaterial::inittextures() {
	textures = new ptrTexture[6];
	for( int i = 0; i < 6; i++ ) {
		textures[i] = NULL;
	}
}

void pMaterial::setDiffuse( rgb d ) {
	diffuse = d;
}

void pMaterial::setSpecular( rgb s) {
	specular = s;
}

void pMaterial::setTexture( int n, ptrTexture tex ) {
	if( n > 0 && n <= 6 ) {
		textures[n-1] = tex;
	}
}

int pMaterial::numTextures() {
	int num = 0;
	if( textures[2] != NULL || textures[3] != NULL || textures[4] != NULL || textures[5] != NULL ){
		num = 6;
	} else if( textures[0] != NULL && textures[1] != NULL ) {
		num = 2;
	} else if( textures[1] != NULL ) {
		num = 6;
	} else if( textures[0] != NULL ) {
		num = 1;
	}
	return num;
}

void pMaterial::renderColour( ) {
	glColor3ubv( (GLubyte*)&diffuse );
}

void pMaterial::getPlane( int num, GLdouble *ps, GLdouble *pt, Point3D &pmin, Point3D &pmax, GLdouble s, GLdouble t, GLdouble r ) {
	double xsize = pmax[0]-pmin[0];
	double ysize = pmax[1]-pmin[1];
	double zsize = pmax[2]-pmin[2];
	double ratio = max( max(xsize, ysize), zsize );
	switch (num) {
		case 2:
			ps[0] = s/ratio; ps[3] = xsize/(2*ratio);
			pt[1] = -t/ratio; pt[3] = ysize/(2*ratio);
			break;
		case 3:
			ps[0] = -s/ratio; ps[3] = xsize/(2*ratio);
			pt[1] = -t/ratio; pt[3] = ysize/(2*ratio);
			break;
		case 0:
			ps[2] = r/ratio; ps[3] = zsize/(2*ratio);
			pt[1] = -t/ratio; pt[3] = ysize/(2*ratio);
			break;
		case 1:
			ps[2] = -r/ratio; ps[3] = zsize/(2*ratio);
			pt[1] = -t/ratio; pt[3] = ysize/(2*ratio);
			break;
		case 4:
			ps[0] = s/ratio; ps[3] = xsize/(2*ratio);
			pt[2] = r/ratio; pt[3] = zsize/(2*ratio);
			break;
		case 5:
			ps[0] = -s/ratio; ps[3] = xsize/(2*ratio);
			pt[2] = r/ratio; pt[3] = zsize/(2*ratio);
			break;
	}
	
}

void pMaterial::renderTextureCube( Point3D &pmin, Point3D &pmax, int num, double s, double t, double r ) {
	if( textures[num] != NULL ) {
		GLdouble planes[4] = {0, 0, 0, 0};
		GLdouble planet[4] = {0, 0, 0, 0};
		getPlane( num, planes, planet, pmin, pmax, s, t, r );
		glBindTexture( GL_TEXTURE_2D, textures[num]->gettex() );
		renderTex( planes, planet );				
	} else {
		glColor3ubv( (GLubyte*)&diffuse );
	}
}

void pMaterial::renderTexture( Point3D &pmin, Point3D &pmax, int num, double s, double t, double r, int axis ) {
	if( numTextures() == 2 ) {
		if( num == 1 ) {
			if( axis % 2 == 0 ) {
				axis++;
			} else {
				axis--;
			}
		}
	}
	GLdouble planes[4] = {0, 0, 0, 0};
	GLdouble planet[4] = {0, 0, 0, 0};
	getPlane( axis, planes, planet, pmin, pmax, s, t, r );
	glBindTexture( GL_TEXTURE_2D, textures[num]->gettex() );
	renderTex( planes, planet );				
}

void pMaterial::renderTex( GLdouble *s, GLdouble *t ) {
		glEnable( GL_TEXTURE_GEN_S );
		glEnable( GL_TEXTURE_GEN_T );
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
		glTexGendv( GL_S, GL_OBJECT_PLANE, s );
		glTexGendv( GL_T, GL_OBJECT_PLANE, t );
}


void pMaterial::clearRender() {
	if( textures[0] != NULL ) {
		glDisable( GL_TEXTURE_GEN_S );
		glDisable( GL_TEXTURE_GEN_T );
	}
}
pMaterial::~pMaterial() {
	delete [] textures;
}

const string& pMaterial::getname() const {
	return name;
}
