/********
Model Class
*********/

#include "model.h"
#include <fstream>
#include <iostream>
#include <sstream>
#include <queue>

static GLuint drawlistcount = 100;

using namespace std;

pModel::pModel( const char *mname ):
name( mname ) {
	drawlist = 0;
	m = NULL;
	taxis = 0;
	axis = Vector3D( 0, 0, 1 );
	tmap = Point3D( 1, 1, 1 );
}

pModel::~pModel(){
	for( unsigned int i = 0; i < f.size(); i++ ) {
		delete f[i];
	}
}

bool pModel::loadOBJ( const char *objname ) {

	return readOBJ( objname );
}

// basic OBJ file reader for vertex and polygon information
bool pModel::readOBJ( const char *objname ) {
	ifstream infile( objname );
	
	int facenum = 0;
	if( ! infile ) return false;
	while( ! infile.eof() ) {
		char c;
		infile.get(c);
		string buf;
		switch( c ) {
			case 'v':
			case 'V':
				{
					float x, y, z;
					infile >> x;
					infile >> y;
					infile >> z;
					v.push_back( Point3D( x, y, z ) );
					//cerr << x << " " << y << " " << z << endl;
					getline( infile, buf );
				} break;
			case 'f':
			case 'F':
				{
					int vert;
					int vertnum = 0;
					infile >> vert;
					
					vector<int> *facev = new vector<int>;
					while( infile ) {
						facev->push_back(vert-1);
						infile >> vert;
						vertnum++;
					}
					if( !facev->empty() ){
						f.push_back( facev );
						facenum++;
					}
 					if( !infile.eof() ) infile.clear();
				} break;
			default :
				{
					if( c != '\n' ) getline( infile, buf );
				}
		}
	}

	infile.close();

	return true;
}

void pModel::setMaterial( pMaterial *mat ) {
	m = mat;
}

// Generate an OpenGL draw list for the model
void pModel::genDrawList( bool smooth ) {
	if( drawlist == 0 ) {
		drawlist = drawlistcount++;
	}
	priority_queue<FaceRef, vector<FaceRef>, PrioritizeFaces> pf;
	int numtex = 0;
	if( m != NULL ) numtex = m->numTextures();
	Vector3D *normals;
	Point3D pmax(-99999, -99999, -99999);
	Point3D pmin(99999, 99999, 99999);
	int *avgnorms;
	for( unsigned int i = 0; i < v.size(); i++ ) {
		if( v[i][0] < pmin[0] ) pmin[0] = v[i][0];
		if( v[i][1] < pmin[1] ) pmin[1] = v[i][1];
		if( v[i][2] < pmin[2] ) pmin[2] = v[i][2];
		if( v[i][0] > pmax[0] ) pmax[0] = v[i][0];
		if( v[i][1] > pmax[1] ) pmax[1] = v[i][1];
		if( v[i][2] > pmax[2] ) pmax[2] = v[i][2];
	}

	// Calculate averaged normals for a smooth surface model
	if( smooth ){
		normals = new Vector3D[ v.size() ];
		avgnorms = new int[ v.size() ];
		for( unsigned int i = 0; i < v.size(); i++ ) {
			avgnorms[i] = 0;
		}
		for( unsigned int i = 0; i < f.size(); i++ ) {
			Point3D p0 = v[(*f[i])[0]];
			Point3D p1 = v[(*f[i])[1]];
			Point3D p2 = v[(*f[i])[2]];
			Vector3D v1 = p1 - p0;
			Vector3D v2 = p1 - p2;
			Vector3D normal = v2.cross( v1 );
			normal.normalize();
			for( unsigned int j = 0; j < f[i]->size(); j++ ) {
				normals[(*f[i])[j]] = normals[(*f[i])[j]] + normal;
				avgnorms[(*f[i])[j]]++;
			}
		}
		for( unsigned int i = 0; i < v.size(); i++ ) {
			normals[i] = ((double)1/avgnorms[i])*normals[i];
		}
	}
	// Calculate which texture to put on the face depending on the polygon normal
	if( numtex > 1 ) {
		for( unsigned int i = 0; i < f.size(); i++ ) {
			Point3D p0 = v[(*f[i])[0]];
			Point3D p1 = v[(*f[i])[1]];
			Point3D p2 = v[(*f[i])[2]];
			Vector3D v1 = p1 - p0;
			Vector3D v2 = p1 - p2;
			Vector3D normal = v2.cross( v1 );
			normal.normalize();
			if( numtex == 2 ) {
				if( normal.dot( axis ) >= 0 ) {
					pf.push( FaceRef( 0, i ) );
				} else {
					pf.push( FaceRef( 1, i ) );
				}
			} else {
				int pri = 0;
				double dotval = normal.dot( Vector3D( 1, 0, 0 ) );
				if( normal.dot( Vector3D( -1, 0, 0 ) ) > dotval ) {
					dotval = normal.dot( Vector3D( -1, 0, 0 ) );
					pri = 1;
				}
				if( normal.dot( Vector3D( 0, 0, 1 ) ) > dotval ) {
					dotval = normal.dot( Vector3D( 0, 0, 1 ) );
					pri = 2;
				} else if( normal.dot( Vector3D( 0, 0, -1 ) ) > dotval ) {
					dotval = normal.dot( Vector3D( 0, 0, -1 ) );
					pri = 3;
				}
				if( normal.dot( Vector3D( 0, 1, 0 ) ) > dotval ) {
					dotval = normal.dot( Vector3D( 0, 1, 0 ) );
					pri = 4;
				} else if( normal.dot( Vector3D( 0, -1, 0 ) ) > dotval ) {
					dotval = normal.dot( Vector3D( 0, -1, 0 ) );
					pri = 5;
				}
				pf.push( FaceRef( pri, i ) );
			}
		}
	} else {
		for( unsigned int i = 0; i < f.size(); i++ ) {
			pf.push( FaceRef( 0, i ) );
		}
	}
	glNewList( drawlist, GL_COMPILE );
	if( m != NULL ) {
		if( numtex == 0 ) {
			m->renderColour( );
		} else if( numtex == 1 ) {
			m->renderTexture( pmin, pmax, 0, tmap[0],tmap[1],tmap[2], taxis );
		}
	}

	int curtex = -1;
	while( !pf.empty() ) {
		FaceRef fr = pf.top();
		pf.pop();
		if( numtex == 2 ) {
			if( curtex < fr.p ) {
				m->renderTexture( pmin, pmax, fr.p, tmap[0],tmap[1],tmap[2], taxis );
				curtex = fr.p;
			}
		} else if( numtex == 6 ) {
			if( curtex < fr.p ) {
				m->renderTextureCube( pmin, pmax, fr.p, tmap[0],tmap[1],tmap[2] );
				curtex = fr.p;
			}
		}
		glBegin( GL_POLYGON );
		if( !smooth ) {
			Vector3D v1 = v[(*f[fr.f])[1]] - v[(*f[fr.f])[2]];
			Vector3D v2 = v[(*f[fr.f])[1]] - v[(*f[fr.f])[0]];
			Vector3D normal = v1.cross( v2 );
			normal.normalize();
			//cerr << "Face: " << i << "  Normal: " << normal << endl;
			for( unsigned int j = 0; j < f[fr.f]->size(); j++ ) {
				glNormal3d( normal[0],normal[1],normal[2]   );
				glVertex3d( v[(*f[fr.f])[j]][0],v[(*f[fr.f])[j]][1],v[(*f[fr.f])[j]][2] );
			}
		} else {
			for( unsigned int j = 0; j < f[fr.f]->size(); j++ ) {
				glNormal3dv( &normals[(*f[fr.f])[j]][0] );
				glVertex3dv( &v[(*f[fr.f])[j]][0] );
			}
		}
		glEnd();
	}
	if( m != NULL ) {
		if( numtex > 0 ) {
			m->clearRender();
		}
	}

	glEndList();
	GLenum error = glGetError();
	if( error != GL_NO_ERROR ) {
		cerr << "Draw List Error" << endl;
	}
	if( smooth ) {
		delete [] normals;
		delete [] avgnorms;
	}
}

void pModel::render( bool rt ) {
	if( drawlist != 0 ) {
		if( !rt && m != NULL ) m->renderColour();
		glCallList( drawlist );
	}
}

const string& pModel::getname() const {
	return name;
}

void pModel::setAxis( int a ) {
	taxis = a;
	switch( a ){
		case 0:
			axis = Vector3D(1,0,0);
			break;
		case 1:
			axis = Vector3D(-1,0,0);
			break;
		case 4:
			axis = Vector3D(0,1,0);
			break;
		case 5:
			axis = Vector3D(0,-1,0);
			break;
		case 2:
			axis = Vector3D(0,0,1);
			break;
		case 3:
			axis = Vector3D(0,0,-1);
			break;
	}
}

void pModel::setTMap( Point3D p ) {
	tmap = p;
}
