/************
Painting Algorithms
*************/

#include "painter.h"
#include "util.h"
#include <cmath>
#include <memory.h>
#include <iostream>

using namespace std;

#define BLUR1 10
#define BLUR2 11
#define BLUR3 12

Painter::Painter(){
	orgimg = NULL;
	lastimg = NULL;
	refimg = NULL;
	imgsat = NULL;
	diff = NULL;
	diffgrid = NULL;
	imgsat = NULL;
	luminance = NULL;
	havetexnames = false;
	brushtexture = false;
	brushtexname = 0;
	start = true;
	D = new Uint16[256];
	Threshold = 25;
	maxlength = 8;
	minlength = 3;
	for( int i = 0; i < 256; i++ ) {
		D[i] = (Uint16)sqrt((double)i*256);
	}

}

Painter::~Painter(){
	if( imgsat != NULL ) delete imgsat;
	if( refimg != NULL ) {
		delete [] refimg;
	}
	if( D != NULL ) {
		delete [] D;
	}
	if(luminance != NULL) {
		delete [] luminance;
	}
	if( diff != NULL ) {
		delete [] diff;
	}
	if( diffgrid != NULL ) {
		delete [] diffgrid;
	}
	if( canvas != NULL ) {
		delete [] canvas;
	}
	if( canvasa != NULL ) {
		delete [] canvasa;
	}

}

void Painter::setbrushtexture( GLuint texname )
{
	//cerr << "texname: " << texname << endl;
	if( texname == 0 ) {
		brushtexture = false;
	} else {
		brushtexture = true;
		brushtexname = texname;
	}
}

void Painter::togletexture()
{
	if( brushtexname != 0 ) {
		brushtexture = !brushtexture;
	}
}

// set the canvas size
void Painter::setsize( int width, int height )
{
	start = true;
	w = width;
	h = height;
	diffw = (w+7)/8;
	diffh = (h+7)/8;
	if( refimg != NULL ) {
		delete [] refimg;
	}
	if( canvas != NULL ) {
		delete [] canvas;
	}
	if( canvasa != NULL ) {
		delete [] canvasa;
	}
	if(screen != NULL) {
		delete [] screen;
	}
	if(luminance != NULL) {
		delete [] luminance;
	}
	if( diff != NULL ) {
		delete [] diff;
	}
	if( diffgrid != NULL ) {
		delete [] diffgrid;
	}
	refimg = new rgb[w*h];
	canvas = new rgb[w*h];
	canvasa = new rgba[w*h];
	luminance = new Uint8[w*h];
	screen = new rgb[w*h];
	diff = new rgbs16[w*h];
	diffgrid = new bool[diffw*diffh];
	memset( diff, 0, sizeof(rgbs16)*w*h );
	if( imgsat != NULL ) {
		delete imgsat;
	}
	if( havetexnames ) {
		glDeleteTextures( NUM_TEXTURES, &texnames[0] );
	}
	glGenTextures( NUM_TEXTURES, &texnames[0] );
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

}

void Painter::paint( rgb *data ) {
	paint( data, NULL, NULL, NULL );
}

void Painter::paint( rgb *data, bool s ) {
	start = s;
	paint( data );
}
void Painter::paint( rgb*data, rgba* blur1, rgba*blur2, rgba*blur3, bool s ) {
	start = s;
	paint( data, blur1, blur2, blur3 );
}
#if 0
// testing for brush strokes
void Painter::paint( rgb*data, rgba* blur1, rgba*blur2, rgba*blur3 )
{
	glPushAttrib( GL_ALL_ATTRIB_BITS );


		glEnable(GL_TEXTURE_2D);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
		glBindTexture(GL_TEXTURE_2D, brushtexname);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glEnable(GL_BLEND);
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		
		glClear(GL_DEPTH_BUFFER_BIT);
		glClear( GL_COLOR_BUFFER_BIT );


	rgb colour;
	colour.r = 255;
	colour.g = 0;
	colour.b = 0;
    BrushStroke bs(32, 0, colour);
	bs.addPoint(200, 50);
	bs.addPoint(200, 100);
	bs.addPoint(200, 150);
	bs.addPoint(200, 200);
	bs.addPoint(200, 250);
	bs.addPoint(200, 300);
	bs.render(true);
	glFlush();
			glPopAttrib();

}
#else
// Painting Algorithm
void Painter::paint( rgb*data, rgba* blur1, rgba*blur2, rgba*blur3 )
{
	orgimg = data;
	//int radius[3] = {4,2,1};
	int radius[4] = {8,4,2};
	rgba *blurs[3];
	blurs[0] = blur3;
	blurs[1] = blur2;
	blurs[2] = blur1;
	Uint32 T = Threshold;
	Uint32 diffT = 2000;
	int count = 0;
	bool ps = true;
	if( blur1 == NULL ) ps = false;

  // Setup brush textures if necessary
	glPushAttrib( GL_ALL_ATTRIB_BITS );
	if( brushtexture ) {
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, brushtexname);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
		glEnable(GL_BLEND);
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	} else {
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);
	}

  // Generated blured image information
	glClear(GL_DEPTH_BUFFER_BIT);
	if( !ps ) {
		if( imgsat == NULL ) {
			imgsat = new SAT( w, h, data );
		} else {
			imgsat->update( data );
		}
	}

	// Copy canvas information to main memory
	memset( diffgrid, 0, diffw*diffh*sizeof(bool));
	if( ps ) {
		getcanvas( canvasa, 0 );
	} else {
		glReadPixels( 0, 0, w, h, GL_RGB, GL_UNSIGNED_BYTE, (void *)canvas);
	}
	// Setup for blurred images in demo mode
	if( ModeDemo && MDstage < 4 ) {
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);
		switch(MDstage){
			case 0:
				glDrawPixels( w, h, GL_RGB, GL_UNSIGNED_BYTE, (void*)orgimg );
				break;
			case 1:
				imgsat->blur( radius[0], refimg );
				glDrawPixels( w, h, GL_RGB, GL_UNSIGNED_BYTE, (void*)refimg );
				break;
			case 2:
				imgsat->blur( radius[1], refimg );
				glDrawPixels( w, h, GL_RGB, GL_UNSIGNED_BYTE, (void*)refimg );
				break;
			case 3:
				imgsat->blur( radius[2], refimg );
				glDrawPixels( w, h, GL_RGB, GL_UNSIGNED_BYTE, (void*)refimg );
				break;
		}
		glPopAttrib();
		lastimg = orgimg;
		return;
				
	}
	
	// Calculate difference information between previous image and current image
	if( !start && (!ModeDemo || lastimg != NULL )) {
		int offset = 0;
		for( int i = 0; i < h; i++ ) {
			for( int j = 0; j < w; j++, offset++ ) {
				if( ModeDemo ) {
					diff[offset].r += (short)orgimg[offset].r - lastimg[offset].r;
					diff[offset].g += (short)orgimg[offset].g - lastimg[offset].g;
					diff[offset].b += (short)orgimg[offset].b - lastimg[offset].b;
				} else {
					if( ps ) { 
						diff[offset].r += (short)orgimg[offset].r - canvasa[offset].r;
						diff[offset].g += (short)orgimg[offset].g - canvasa[offset].g;
						diff[offset].b += (short)orgimg[offset].b - canvasa[offset].b;
					} else {
						diff[offset].r += (short)orgimg[offset].r - lastimg[offset].r;
						diff[offset].g += (short)orgimg[offset].g - lastimg[offset].g;
						diff[offset].b += (short)orgimg[offset].b - lastimg[offset].b;
					}
				}
			}
		}
		// Go through the image and find which grid points are in need of new brush strokes
		// based on animation frame differences
		for( int i = 0; i < h; i+=8 ) {
			for( int j = 0; j < w; j+=8 ) {
				Uint32 sum = 0;
				for( int k = i; k < i + 8; k++ ) { 
					for( int l = j; l < j + 8; l++ ) {
						sum += abs(diff[ k*w + l].r)+abs(diff[ k*w + l].g)
							+abs(diff[ k*w + l].b);
					}
				}
				if( sum > diffT ) {
					count++;
					for( int k = i; k < i + 8; k++ ) { 
						for( int l = j; l < j + 8; l++ ) {
							diff[k*w+l].r = 0;
							diff[k*w+l].g = 0;
							diff[k*w+l].b = 0;
						}
					}
					diffgrid[(i/8)*diffw+(j/8)] = true;
				} else {
					diffgrid[(i/8)*diffw+(j/8)] = false;
				}
			}
		}
	}
	// setup which brush sizes to draw for demo mode
	int qstart = 0;
	int qend = 3;
	if( ModeDemo ) {
		if( MDstage == 4 || MDstage == 7 || MDstage == 8 || MDstage == 9 ) {
			memset( canvas, 255, w*h*3 );
			glClear( GL_COLOR_BUFFER_BIT );
			qstart = 0; qend = 1;
		} else if( MDstage == 5 ) {
			if( MDstageprev == 4 ) {
				qstart = 1; qend = 2;
			} else {
				memset( canvas, 255, w*h*3 );
				glClear( GL_COLOR_BUFFER_BIT );
				qstart = 0; qend = 2;
			}
		} else if( MDstage == 6 ) {
			if( MDstageprev == 5 ) {
				qstart = 2; qend = 3;
			} else {
				memset( canvas, 255, w*h*3 );
				glClear( GL_COLOR_BUFFER_BIT );
			}
		}
	}
	
	// generate the brush strokes
	for( int q = qstart; q < qend; q++ ) {
		if( ps ) {
			refimga = blurs[q];
		} else {
			imgsat->blur( radius[q], refimg );
		}
		glClear( GL_DEPTH_BUFFER_BIT );
		// get updated canvas
		if( q > 0 ) {
			if( ps ) {
				getcanvas( canvasa, q );
			} else { 
				glReadPixels( 0, 0, w, h, GL_RGB, GL_UNSIGNED_BYTE, (void *)canvas);
			}
		}
		// calculate luminance of the image for gradient detection
		int offset = 0;
		if( !ps ) {
			for( int i = 0; i < h; i++ ) {
				for( int j = 0; j < w; j++, offset++ ) {
					luminance[offset] = (Uint8)(0.3f*refimg[offset].r +
										0.59f*refimg[offset].g +
										0.11f*refimg[offset].b);
				}
			}
		} 
		int grid = radius[q];
		double z = -254;
		int oset = 0;
		count = 0;
		// loop through grid points to generate new strokes
		for( int y = 0; y+grid < h; y += grid ) {
			for( int x = 0; x+grid < w; x += grid, oset++ ) {
				if( start || diffgrid[ (y/8)*diffw + (x/8)] ) {
					count++;
					Uint16 maxval = 0;
					Uint16 edelta = 0;
					int mx, my;
					Uint32 error = 0;
					
					//calculate the difference between the blured image and the canvas to see if a stroke is necessary
					for( int i = y; i < y+grid; i++ ) {
						int offset = i*w + x;
						for( int j = x; j < x+grid; j++, offset++ ) {
							if( ps ) {
								edelta = D[canvasa[offset].a];
							} else {
								edelta = (Uint16)sqrt((double)(canvas[offset].r-refimg[offset].r)*(canvas[offset].r-refimg[offset].r)
										+ (canvas[offset].g-refimg[offset].g)*(canvas[offset].g-refimg[offset].g)
										+ (canvas[offset].b-refimg[offset].b)*(canvas[offset].b-refimg[offset].b));
							}
							error += edelta;
							if( edelta > maxval ) {
								maxval = edelta;
								mx = j;
								my = i;
							}
						}
					}
					// if the differences is big enough create a stroke
					if( error / (grid * grid) > T ) {
						float z = -((float)rand()/RAND_MAX);
						if( ModeDemo && ( MDstage == 7 || MDstage == 8 || MDstage == 9 ) ) {
							if( oset % 73 == 0 ) {
								strokes.push( Point( mx, my, z ) );
							}
						} else {
							strokes.push( Point( mx, my, z ) );
						}
					}
				}
			}
		}
		// generate the complete brush strokes
		while( !strokes.empty() ) {
			Point p = strokes.top();
			strokes.pop();
			genbrushstroke( (int)p.x, (int)p.y, p.z, radius[q], ps );
		}
	}
		glFlush();
	lastimg = orgimg;
	start = false;
	glPopAttrib( );
}
#endif

void inline Painter::drawdot( int x, int y, double z, int R, rgb colour ) {
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glPushMatrix();
	glTranslated( x, y, z );
	glColor3ub( colour.r, colour.g, colour.b );
	glCallList(R); 
	glPopMatrix();
	
}

#define RGBDIFF(x,y) (abs(x.r-y.r)+abs(x.g-y.g)+abs(x.b-y.b))

// generate a brush stroke based on a starting position
void Painter::genbrushstroke( int x, int y, float z, int R, bool ps )
{
	int offset = 0;
		
	rgb colour;
	// find the colour for the stroke
	if( ps ) {
		colour.r = refimga[y*w+x].r;	
		colour.g = refimga[y*w+x].g;
		colour.b = refimga[y*w+x].b;
	} else {
		colour = refimg[y*w+x];
	}
	double lastdx, lastdy;
	lastdx = 0;
	lastdy = 0;
	BrushStroke bs(R, 0, colour);

	if( ModeDemo && MDstage == 7 ) drawdot( x, y, z, 2, colour );
	bs.addPoint(x, y);
	// generate points until the max length of the stroke is reached
	for( int i = 0; i < maxlength; i++ ) {
		// make sure the points are inside the border of the iamge
		if( x == 0 ) x = 1;
		else if( x == w-1 ) x = w-2;
		if( y == 0 ) y = 1;
		else if( y == h-1 ) y = h-2;

		offset = y*w+x;
		if( i > 0 ) {
			if( x < 0 || y < 0 || x >= w || y >= h ) break;
		}
		// update the difference grid when new strokes are made
		if( !start && !diffgrid[ (y/8)*diffw+(x/8) ] ) {
			diffgrid[ (y/8)*diffw+(x/8) ] = true;
			break;
		}
		// if the minimum stroke length is reached and the colour is different than the original image at this
		// point, stop drawing the stroke
		if (i > minlength ) {
			Uint16 diff1;
			Uint16 diff2;
			if( ps ) {
				diff1 = RGBDIFF(refimga[offset],canvasa[offset]);
				diff2 = RGBDIFF(refimga[offset],colour);
			} else {
				diff1 = RGBDIFF(refimg[offset],canvas[offset]);
				diff2 = RGBDIFF(refimg[offset],colour);
			}
			if( diff1 < diff2 )
				break;
		}
		// calculate the gradient vector
		short Gh, Gv;
		if( ps ) {
			Gv = (short)refimga[offset+w-1].a - (short)refimga[offset-w-1].a +
					2*((short)refimga[offset+w].a - (short)refimga[offset-w].a) +
					(short)refimga[offset+w+1].a - (short)refimga[offset-w+1].a;
			Gh = (short)refimga[offset-w+1].a - (short)refimga[offset-w-1].a +
					2*((short)refimga[offset+1].a - (short)refimga[offset-1].a) +
					(short)refimga[offset-w+1].a - (short)refimga[offset-w-1].a;
		} else {
			Gv = luminance[offset+w-1] - luminance[offset-w-1] +
					2*(luminance[offset+w] - luminance[offset-w]) +
					luminance[offset+w+1] - luminance[offset-w+1];
			Gh = luminance[offset-w+1] - luminance[offset-w-1] +
					2*(luminance[offset+1] - luminance[offset-1]) +
					luminance[offset-w+1] - luminance[offset-w-1];
		}
		if( abs(Gh + Gv) == 0 ) break;
		double dx = Gv;
		double dy = -Gh;


	  // generate new stroke point based on gradient vector	
		double denom = sqrt(dx*dx + dy*dy);
		if( denom != 0 ){
			dx = dx/denom;
			dy = dy/denom;
		}
		if ( lastdx*dx + lastdy*dy < 0 ) {
			dx = -dx;
			dy = -dy;
		}
		x = x+(int)(R*dx);
		y = y+(int)(R*dy);
		lastdx = dx;
		lastdy = dy;
		if( ModeDemo && MDstage == 7 ) drawdot( x, y, z, 2, colour );
		if( !start && x >= 0 && y >= 0 && x < w && y < h && !diffgrid[ (y/8)*diffw+(x/8) ] ) {
			diffgrid[ (y/8)*diffw+(x/8) ] = true;
			break;
		}

		bs.addPoint(x, y);

	}
	if( !ModeDemo || MDstage != 7) bs.render( brushtexture );

}

void Painter::setstrokeparam( Uint32 T, int maxlen, int minlen ) {
	Threshold = T;
	maxlength = maxlen;
	minlength = minlen;
}
