/************
Painting Algorithms
*************/

#ifndef __PAINTER_H__
#define __PAINTER_H__

#include "sat.h"
#include "stroke.h"
#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#define SCREEN_DEPTH 24
#define NUM_TEXTURES 6
#include <vector>
#include <queue>

class Painter {
public:
	Painter();
	~Painter();

	void paint( rgb *data, rgba* blur1, rgba*blur2, rgba*blur3, bool s );
	void paint( rgb *data, bool s );
	void paint( rgb *data, rgba* blur1, rgba*blur2, rgba*blur3 );
	void paint( rgb *data );

	void genbrushstroke( int x, int y, float z, int R, bool ps );
	void setstrokeparam( Uint32 T, int maxlen, int minlen );

	void setsize( int width, int height );
	void setbrushtexture( GLuint texname );
	void togletexture();

private:
	void inline drawdot( int x, int y, double z, int R, rgb colour );

	rgb *canvas;
	rgba *canvasa;
	rgb *orgimg;
	rgb *lastimg;
	rgb *refimg;
	rgba *refimga;
	rgb *screen;
	rgbs16 *diff;
	Uint8 *luminance;
	int h, w;
	SAT *imgsat;
	Uint16* D;
	GLuint texnames[NUM_TEXTURES];
	bool havetexnames;
	GLuint brushtexname;
	bool brushtexture;
	priority_queue<Point, vector<Point>, PrioritizePoints> strokes;
	bool start;
	bool *diffgrid;
	int diffw, diffh;
	int maxlength, minlength;
	Uint32 Threshold;

};

#endif
