/*******************************
Particle System Class
*******************************/
#include "particle.h"
#include <queue>

Particle::Particle(){
	origin = Point3D(0,0,0);
	velocity = Vector3D(0,0,0);
	color = RGB_0;
	start = 0;
	end = 0;
	accel = 0;
}

Particle::~Particle(){

}

void Particle::initialize( Point3D &org, Vector3D &vel, double a, int s, int life ){
	origin = org;
	velocity = vel;
	dir = velocity;
	dir.normalize();
	accel = a;
	start = s;
	end = s + life;
	
}

Point3D Particle::pos( int tick ) {
	double time = (double)(tick-start)/1000;
	return origin + time*velocity + (time*time)*accel/2*dir;
}


void Particle::render( int tick ) {
	if( tick <= end ) {
		double time = (double)(end-tick)/1000;
		double etime = (double)(end-start)/1000;
		double alpha =  (time - (time*time)/200)/(etime - (etime*etime)/200);
		Point3D o = pos(tick);
		glPushMatrix();
		glColor4d( (double)color.r/255, (double)color.g/255, (double)color.b/255, alpha );
		glTranslated( o[0], o[1], o[2] );
		glCallList( particlelist );

		glPopMatrix();
	}
}

ParticleSystem::ParticleSystem( int s ) {
	start = s;
	middle = s + 3000;
	end = middle + 4000;
	group1 = 5;
	group2 = 60;
}

ParticleSystem::~ParticleSystem() {
	while( !p1.empty() ) {
		Particle * p = p1.back();
		p1.pop_back();
		delete p;
	}
	while( !p2.empty() ) {
		Particle * p = p2.back();
		p2.pop_back();
		delete p;
	}
}

void ParticleSystem::initialize( Point3D &org, Vector3D &vel, double a, rgb color ){
	double r = vel.length();
	double a2 = atan( vel[1]/vel[0] );
	double a1 = acos( vel[2]/r );
	for( int i = 0; i < group1; i++ ) {
		double da1 = deg2rad(20.0 - 40.0*((double)rand()/RAND_MAX));
		double da2 = deg2rad(20.0 - 40.0*((double)rand()/RAND_MAX));
		Vector3D pv(r*sin(a1+da1)*cos(a2+da2),
					r*sin(a1+da1)*sin(a2+da2),
					r*cos(a1+da1) );
		Particle * p = new Particle();
		p->setColor( color );
		p->initialize( org, pv, a, start, middle-start );
		p1.push_back( p );
		for( int j = 0; j < group2; j++ ) {
			double ra1 = deg2rad(360.0*((double)rand()/RAND_MAX));
			double ra2 = deg2rad(360.0*((double)rand()/RAND_MAX));
			double rr = (1.2 - 0.4*(double)rand()/RAND_MAX)*r;
			double len = 1 - 0.4*(double)rand()/RAND_MAX;
			Vector3D ppv(rr*sin(ra1)*cos(ra2),
						rr*sin(ra1)*sin(ra2),
						rr*cos(ra1) );
			Particle * pp = new Particle();
			pp->setColor( color );
			Point3D ppoint = p->pos(middle);
			pp->initialize( ppoint, ppv, a, middle, (int)((end-middle)*len) );
			p2.push_back( pp );
		}
	}
}

void ParticleSystem::render( int tick, prq &drawp ) {
	PRef pr;
	if( tick < middle && tick >= start) {
		for( unsigned int i = 0; i < p1.size(); i++ ) {
			pr.p = p1[i];
			pr.z = p1[i]->pos(tick)[2];
			drawp.push( pr );
		}
	} else if( tick <= end ) {
		for( unsigned int i = 0; i < p2.size(); i++ ) {
			pr.p = p2[i];
			pr.z = p2[i]->pos(tick)[2];
			drawp.push( pr );
		}
	}

}
