#ifndef __PARTICLE_H__
#define __PARTICLE_H__

#include "util.h"
#include "algebra.h"
#include <vector>
#include <queue>

using namespace std;

class Particle {
public:
	Particle();
	~Particle();
	void initialize( Point3D &org, Vector3D &vel, double a, int s, int life );
	void setColor( rgb c ) { color = c;};

	void render( int tick );
	Point3D pos( int tick );
private:
	Point3D origin;
	Vector3D velocity;
	Vector3D dir;
	rgb color;
	int start, end;
	double accel;

};

class PRef {
public:
	PRef( ) { z = 0; p = NULL;};
	~PRef(){};
	double z;
	Particle *p;
};

class PrioritizeParticles { 
public: 
	int operator()( const PRef &a, const PRef &b ) 
	{ return a.z > b.z; };
};

typedef priority_queue< PRef, vector<PRef>, PrioritizeParticles > prq;

class ParticleSystem {
public:
	ParticleSystem( int s );
	~ParticleSystem();

	void initialize( Point3D &org, Vector3D &vel, double a, rgb color );

	void render( int tick, prq &drawp );
	int getStart() { return start; };
	int getEnd() { return end; };

private:
	int start, middle, end;
	int group1, group2;
	vector< Particle* > p1;
	vector< Particle* > p2;
};

#endif
