/***************
Summed Area Table Class
*****************/

#include "sat.h"
#include <iostream>
using namespace std;

// Generate the SAT from an SDL surface
SAT::SAT( SDL_Surface * image ) {
	height = image->h+1;
	width = image->w+1;
	data = new RGB32[width*height];
	RGB32 *rgbdata = data;
	Uint32 offset = 0;
	SDL_LockSurface( image );
	for( int i = 0; i < height; i++ ) {
		Uint8 *imgdata = (Uint8 *)image->pixels;
		if( i > 0 ) imgdata += (i-1) * image->pitch;
		RGB32 sum;
		sum.r = 0;
		sum.g = 0;
		sum.b = 0;
		for( int j = 0; j < width; j++, offset++ ) {
			if( i == 0 || j == 0) {
				rgbdata[offset].r = 0;
				rgbdata[offset].g = 0;
				rgbdata[offset].b = 0;
			} else {
				rgb colour;
				SDL_GetRGB( *(Uint32*)imgdata, image->format, &colour.r, &colour.g, &colour.b );
				sum.r += colour.r;
				sum.g += colour.g;
				sum.b += colour.b;
				rgbdata[offset].r = sum.r;
				rgbdata[offset].g = sum.g;
				rgbdata[offset].b = sum.b;
				if( j > 0 ) {
					rgbdata[offset].r += rgbdata[offset-width].r;
					rgbdata[offset].g += rgbdata[offset-width].g;
					rgbdata[offset].b += rgbdata[offset-width].b;
					imgdata += image->format->BytesPerPixel;
				}
				//cerr << "(" << (Uint32)(*imgdata).r << "," << (Uint32)(*imgdata).g << "," << (Uint32)(*imgdata).b << ")" << endl;
			}
			//cerr << "(" << rgbdata[offset].r << "," << rgbdata[offset].g << "," << rgbdata[offset].b << ")" << endl;
					
		}
	}
	SDL_UnlockSurface( image );

}

SAT::SAT( int w, int h, rgb* image ) {
	height = h+1;
	width = w+1;
	data = new RGB32[width*height];
	update( image );
}

SAT::~SAT() {
	delete [] data;
}

// Generate SAT from raw image data
void SAT::update( rgb* image ) {
	RGB32 *rgbdata = data;
	Uint32 offset = 0;
	for( int i = 0; i < height; i++ ) {
		RGB32 sum;
		sum.r = 0;
		sum.g = 0;
		sum.b = 0;
		for( int j = 0; j < width; j++, offset++ ) {
			if( (i == 0) || (j == 0)) {
				rgbdata[offset].r = 0;
				rgbdata[offset].g = 0;
				rgbdata[offset].b = 0;
			} else {
				sum.r += (*image).r;
				sum.g += (*image).g;
				sum.b += (*image).b;
				rgbdata[offset].r = sum.r + rgbdata[offset-width].r;
				rgbdata[offset].g = sum.g + rgbdata[offset-width].g;
				rgbdata[offset].b = sum.b + rgbdata[offset-width].b;
				image++;
			}
		}
	}
	
}

//generate linear blur
void SAT::blur( int radius, void *dst ) const {
	rgb *rgbdst = (rgb*)dst;
	int xbuf = width-radius;
	int ybuf = height-radius;
	int x1, y1, x2, y2;
	int k, l;
	k=1;
	for( int i = 0; i < height-1; i++, k++) {
		y1 = (i >= radius)?(i-radius):0;
		y2 = (k < ybuf)?(k+radius):height-1;
		l = 1;
		for( int j = 0; j < width-1; j++, l++, rgbdst++ ) {
			x1 = (j >= radius)?(j-radius):0;
			x2 = (l < xbuf)?(l+radius):width-1;
			*rgbdst = _rect( x1, y1, x2, y2 );
			//cerr << "(" << (Uint32)colour.r << "," << (Uint32)colour.g << "," << (Uint32)colour.b << ")" << endl;
		}
	}

}

// generate linear blur
void SAT::blur( int radius, SDL_Surface *image ) const {
	int xbuf = width-radius;
	int ybuf = height-radius;
	int x1, y1, x2, y2;
	int k, l;
	k=1;
	SDL_LockSurface( image );
	for( int i = 0; i < height-1; i++, k++) {
		Uint8 *imgdata = (Uint8 *)image->pixels;
		imgdata += i * image->pitch;
		y1 = (i >= radius)?(i-radius):0;
		y2 = (k < ybuf)?(k+radius):height-1;
		l = 1;
		for( int j = 0; j < width-1; j++, l++ ) {
			x1 = (j >= radius)?(j-radius):0;
			x2 = (l < xbuf)?(l+radius):width-1;
			rgb colour = _rect( x1, y1, x2, y2 );
			*(Uint32*)imgdata = SDL_MapRGB( image->format, colour.r, colour.g, colour.b );
			imgdata += image->format->BytesPerPixel;
			//cerr << "(" << (Uint32)colour.r << "," << (Uint32)colour.g << "," << (Uint32)colour.b << ")" << endl;
		}
	}
	SDL_UnlockSurface( image );

}

rgb SAT::rect( int x1, int y1, int x2, int y2 ) const {
	if( x1 >= 0 && x2 >= 0 && y1 >= 0 && y2 >= 0
		&& x1 < width && x2 < width && y1 < height && y2 < height 
		&& (x1 != x2) && (y1 != y2 ) ){
			return _rect( x1, y1, x2, y2 );
	} else {
		rgb ret;
		ret.r = 0;
		ret.g = 0;
		ret.b = 0;
		return ret;
	}
}

rgb SAT::_rect( int x1, int y1, int x2, int y2 ) const {
	rgb ret;
	//cerr << "(" << x1 << "," << y1 << "), (" << x2 << "," << y2 << ")" << endl;
	ret.r = (Uint8)((data[ y1*width + x1 ].r + data[ y2*width + x2 ].r -
		data[ y2*width + x1 ].r - data[ y1*width + x2 ].r)/((x2-x1)*(y2-y1))); 
	ret.g = (Uint8)((data[ x1 + y1*width ].g + data[x2 + y2*width].g -
		data[ x1 + y2*width].g - data[ x2 + y1*width].g)/((x2-x1)*(y2-y1))); 
	ret.b = (Uint8)((data[ x1 + y1*width ].b + data[x2 + y2*width].b -
		data[ x1 + y2*width].b - data[ x2 + y1*width].b)/((x2-x1)*(y2-y1))); 
	//cerr << "  " << (Uint32)data[ x1 + y1*width ].b << " + " << (Uint32)data[x2 + y2*width].b <<
	//	" - " << (Uint32)data[ x1 + y2*width].b << " - " << (Uint32)data[ x2 + y1*width].b <<
	//	" / " << ((x2-x1)*(y2-y1)) << endl;
	return ret;
}
