/********
Brush Stroke Class
*********/
#include "stroke.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


BrushStroke::BrushStroke( int radius, double z, rgb colour )
{
	_subdivided = false;
	_radius = radius;
	_z = z;
	_colour = colour;
	_subdivision = NULL;
	_temp = NULL;
}

BrushStroke::~BrushStroke()
{
	if( _subdivision != NULL ) {
		delete _subdivision;
	}
	if( _temp != NULL ) {
		delete _temp;
	}
}

void BrushStroke::addPoint( int x, int y )
{
	_controls.push_back( Point(x, y) );
}

void BrushStroke::render( bool texture )
{
	_texture = texture;
	if( !_subdivided ) {
		_subdivision = new vector<Point>;
		subdivide( &_controls, _subdivision );
	}
	drawstroke( _subdivision );
}

float BrushStroke::strokelength( vector<Point> *curve, vector<float> *lengths )
{
	float length = (float)2*_radius;
	Point p0, p1;

	for( Uint8 i = 0; i < curve->size()-1; i++ ) {
		p0 = (*curve)[i];
		p1 = (*curve)[i+1];
		float dist = p0.dist(p1);
		lengths->push_back(dist);
		length+=dist;
	}
	return length;
}

float BrushStroke::strokelength( vector<Point> *curve )
{
	float length = (float)2*_radius;
	Point p0, p1;

	for( Uint8 i = 0; i < curve->size()-1; i++ ) {
		p0 = (*curve)[i];
		p1 = (*curve)[i+1];
		length+=p0.dist(p1);
	}
	return length;
}

void BrushStroke::subdivide( vector<Point> *in, vector<Point> *out )
{
	out->erase( out->begin(), out->end() );
	
	Point p0;
	Point p1;
	Point p2;

	p0 = (*in)[0];

	_subdivided = true;

	out->push_back( Point( p0.x, p0.y ) );

	if( in->size() == 1 ) return;

	p1 = (*in)[1];

	if( in->size() == 2 ) {
		out->push_back( Point( p1.x, p1.y ) );
		return;
	}
	
	out->push_back( Point( (p1.x+p0.x)/2, (p1.y+p0.y)/2 ) );

	for( Uint8 i = 1; i < in->size()-1; i++ ) {
		p0 = (*in)[i-1];
		p1 = (*in)[i];
		p2 = (*in)[i+1];

		out->push_back( Point( (p0.x + 4*p1.x + p2.x)/6,
								(p0.y + 4*p1.y + p2.y)/6 ) );

		out->push_back( Point( (p2.x+p1.x)/2, (p2.y+p1.y)/2 ) );
	}

	out->push_back( Point( p2.x, p2.y ) );
}

void inline BrushStroke::drawdot( Point p ) {
	drawcap( p, 1, 0, true, 2.0*_radius );
	drawcap( p, -1, 0, false, 2.0*_radius );
}

void BrushStroke::drawstroke( vector<Point> *curve ) {

	if( curve->empty() ) return;

	glColor3ubv( (GLubyte *)&_colour );
	if( ModeDemo && MDstage == 8 )
		glColor3f( (float)rand()/RAND_MAX, (float)rand()/RAND_MAX, (float)rand()/RAND_MAX);
	float dx, dy, mag;
	Point p0, p1, p2;
	float length, dlen, ratio, TexU;
	length = 0;

	p0 = (*curve)[0];
	if( curve->size() == 1 ) {
		drawdot( p0 );
		return;
	}

	if( _texture ) {
		length = strokelength( curve );
		dlen = (float)(length - 2*_radius)/(curve->size() - 1)/length;
		ratio = 1/length;
		TexU = 0;
	}
	//cerr << "length: " << length << endl;

	p1 = (*curve)[1];

	dx = p1.y - p0.y;
	dy = p0.x - p1.x;

	mag = sqrtf( dx*dx + dy*dy );

	if( mag != 0 ){
		dx /= mag;
		dy /= mag;
	}
	drawcap(p0, dx, dy, true, length);

	if( _texture ) {
		TexU = _radius/length;
	}
	glBegin( GL_TRIANGLE_STRIP );

	for( Uint8 i = 1; i < curve->size()-1; i++ ) {
		if( _texture ) TexU = ((float)(i-1) /(float)(curve->size()-1))*(length - 2*_radius)/length + _radius/length;
		p0 = (*curve)[i-1];
		p2 = (*curve)[i+1];

		dx = p2.y - p0.y;
		dy = p0.x - p2.x;

		mag = sqrtf( dx*dx + dy*dy );

		if( mag != 0 ){
			dx /= mag;
			dy /= mag;
		}

		if( ModeDemo && MDstage == 8 )
			glColor3f( (float)rand()/RAND_MAX, (float)rand()/RAND_MAX, (float)rand()/RAND_MAX);
		if( _texture ) glTexCoord2f( 0, TexU);
		glVertex3d(p0.x + _radius*dx, p0.y + _radius*dy, _z);
	
		if( _texture ) glTexCoord2f( 1, TexU);
		glVertex3d(p0.x - _radius*dx, p0.y - _radius*dy, _z);
	}
	if( _texture ) TexU = (length-_radius)/length;
	p0 = (*curve)[curve->size()-2];
	p1 = (*curve)[curve->size()-1];

	dx = p1.y - p0.y;
	dy = p0.x - p1.x;

	mag = sqrtf( dx*dx + dy*dy );

	if( mag != 0 ){
		dx /= mag;
		dy /= mag;
	}

	if( _texture ) glTexCoord2f( 0, TexU);
	glVertex3d(p1.x + _radius*dx, p1.y + _radius*dy, _z);
	if( _texture ) glTexCoord2f( 1, TexU);
	glVertex3d(p1.x - _radius*dx, p1.y - _radius*dy, _z);

	glEnd();
	drawcap(p1, -dx, -dy, false, length);
	//drawdot( p1 );

}

void BrushStroke::drawcap( Point p, float dx, float dy, bool start, float length )
{
	double theta = atan2((double)-dy,(double)-dx);
	float TexU, ratio;
	if( _texture ) {
		if( start ) {
			TexU = (float)_radius;
		} else {
			TexU = (float)length-_radius;
		}
		ratio = 1/length;
	}
	//cerr << "TexU: " << (TexU*ratio) << endl;

	glPushMatrix();
	glTranslated(p.x,p.y,_z);

	glBegin(GL_TRIANGLE_FAN);

	if( _texture ) glTexCoord2f(.5f, TexU*ratio);
	glVertex3i(0,0,0);
	
	double textheta = atan2( 1.0, 0.0 );

	for(int i=0;i<=17;i++) {
		double dx1 = cos(theta + i * M_PI / 17);
		double dy1 = sin(theta + i * M_PI / 17);
		if( _texture ) {
			
			double tdx = cos(i * M_PI / 17 );
			double tdy = sin(i * M_PI / 17 );
			
			if(start ) {
				//cerr << "X: " << (0.5 * tdx ) + 0.5 << "  Y: " << (TexU - (_radius * tdy))*ratio << endl;
				glTexCoord2d( (0.5 * tdx ) + 0.5 , (TexU - (_radius * tdy))*ratio );
			} else {
				//cerr << "X: " << 0.5 - (0.5 * tdx ) << "  Y: " << (TexU + (_radius * tdy))*ratio << endl;
				glTexCoord2d( 0.5 - (0.5 * tdx ) , (TexU + (_radius * tdy))*ratio );
			}
		}
		//cerr << "X: " << (_radius * dx1) << "  Y: " << (_radius * dy1) << endl;
		glVertex3i((int)(_radius * dx1), (int)(_radius * dy1), 0);
	}

	glEnd();
	glPopMatrix();
}
