/********
Brush Stroke Class
*********/

#ifndef __STROKE_H__
#define __STROKE_H__

#include <vector>
#include "util.h"
#include <cmath>
#include "sat.h"
#include <iostream>

#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

class Point {
public:
	friend class PrioritizePoints;
	Point():x(0),y(0),z(0){}
	Point( float x, float y ):x(x),y(y),z(0){}
	Point( int x, int y ):x((float)x),y((float)y),z(0){}
	Point( float x, float y, float z ):x(x),y(y),z(z){}
	Point( int x, int y, float z ):x((float)x),y((float)y),z((float)z){}
	inline float dist( Point p ) {
		return sqrtf( (x-p.x)*(x-p.x)+(y-p.y)*(y-p.y) );
	}
	float x, y, z;
};

class PrioritizePoints { 
public: 
	int operator()( const Point &a, const Point &b ) 
	{ return a.z < b.z; } 
};

class BrushStroke {
public:
	BrushStroke( int radius, double z, rgb colour );
	~BrushStroke();

	void addPoint( int x, int y );
	void render( bool texture );

private:
	void subdivide( vector<Point> *in, vector<Point> *out );
	void drawstroke( vector<Point> *curve );
	void drawcap( Point p, float dx, float dy, bool start, float length );
	inline void drawdot( Point p );
	float strokelength( vector<Point> *curve, vector<float> *lengths );
	float strokelength( vector<Point> *curve );

	vector<Point> _controls;
	vector<Point> *_subdivision;
	vector<Point> *_temp;
	
	bool _subdivided;
	int _radius;
	double _z;
	bool _texture;
	rgb _colour;
};



#endif //__STROKE_H__
